<?php

namespace Drupal\webform_googlesheets\Plugin\WebformHandler;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\webform\Plugin\WebformElement\WebformManagedFileBase;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformSubmissionInterface;
use Exception;
use Google\Service\Sheets;
use Google\Service\Sheets\AppendCellsRequest;
use Google\Service\Sheets\BatchUpdateSpreadsheetRequest;
use Google\Service\Sheets\CellData;
use Google\Service\Sheets\ExtendedValue;
use Google\Service\Sheets\Request;
use Google\Service\Sheets\RowData;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form submission to Google Sheets handler.
 *
 * @WebformHandler(
 *   id = "googlesheets",
 *   label = @Translation("Google Sheets"),
 *   category = @Translation("Google Sheets"),
 *   description = @Translation("Adds a form submission to a Google Sheets."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_OPTIONAL,
 *   tokens = TRUE,
 * )
 */
class WebformGoogleSheetsHandler extends WebformHandlerBase {

  /**
   * A webform element plugin manager.
   *
   * @var WebformElementManagerInterface
   */
  protected $elementManager;

  /**
   * Google API Client.
   *
   * @var GoogleApiClientService
   */
  protected $googleApiClient;

  /**
   * Google API Service Client.
   *
   * @var GoogleApiServiceClientService
   */
  protected $googleApiServiceClient;

  /**
   * List of unsupported webform submission properties.
   *
   * The below properties will not being included in a remote post.
   *
   * @var array
   */
  protected $unsupportedProperties = [
    'metatag',
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->elementManager = $container->get('plugin.manager.webform.element');
    $instance->googleApiClient = $container->get('google_api_client.client');
    $instance->googleApiServiceClient = $container->get('google_api_service_client.client');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $configuration = $this->getConfiguration();

    return [
      '#settings' => $configuration['settings'],
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $field_names = array_keys(
      \Drupal::service('entity_field.manager')
        ->getBaseFieldDefinitions('webform_submission')
    );

    return [
      'spreadsheet_url' => '',
      'spreadsheet_id' => '',
      'spreadsheet_sheet_id' => '',
      'google_api_credential_type' => 'google_api_client',
      'google_api_client' => 0,
      'google_api_service_client' => '',
      'multivalue_limit' => 0,
      'multivalue_empty' => 0,
      'sort_data' => 0,
      'included_data' => array_combine($field_names, $field_names),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    parent::buildConfigurationForm($form, $form_state);

    // Google Sheets configuration.
    $form['googlesheets'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Google API settings'),
    ];
    $form['googlesheets']['spreadsheet_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Sheet URL'),
      '#required' => TRUE,
      '#default_value' => $this->configuration['spreadsheet_url'] ?? '',
      '#empty_value' => '',
    ];

    $form['googlesheets']['google_api_credential_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Credential Type'),
      '#required' => TRUE,
      '#options' => [
        'google_api_client' => $this->t('OAuth 2.0 Client'),
        'google_api_service_client' => $this->t('Service Account'),
      ],
      '#default_value' => $this->configuration['google_api_credential_type'] ?? 'google_api_client',
    ];

    $form['googlesheets']['google_api_client'] = [
      '#type' => 'select',
      '#title' => $this->t('OAuth 2.0 Client'),
      '#options' => [],
      '#default_value' => $this->configuration['google_api_client'] ?? 0,
      '#empty_value' => 0,
      '#states' => [
        'visible' => array(
          ':input[name="settings[google_api_credential_type]"]' => ['value' => 'google_api_client'],
        ),
        'required' => array(
          ':input[name="settings[google_api_credential_type]"]' => ['value' => 'google_api_client'],
        ),
      ],
    ];
    $google_api_clients = $this->getEntityStorage('google_api_client')->loadMultiple();
    foreach ($google_api_clients as $google_api_client) {
      $form['googlesheets']['google_api_client']['#options'][$google_api_client->id()] = $google_api_client->label();
    }

    $form['googlesheets']['google_api_service_client'] = [
      '#type' => 'select',
      '#title' => $this->t('Service Account'),
      '#options' => [],
      '#default_value' => $this->configuration['google_api_service_client'] ?? '',
      '#empty_value' => '',
      '#states' => [
        'visible' => array(
          ':input[name="settings[google_api_credential_type]"]' => ['value' => 'google_api_service_client'],
        ),
        'required' => array(
          ':input[name="settings[google_api_credential_type]"]' => ['value' => 'google_api_service_client'],
        ),
      ],
    ];
    $google_api_service_clients = $this->getEntityStorage('google_api_service_client')->loadMultiple();
    foreach ($google_api_service_clients as $google_api_service_client) {
      $form['googlesheets']['google_api_service_client']['#options'][$google_api_service_client->id()] = $google_api_service_client->label();
    }

    // Prepare to check for potential issues.
    $label = '';
    $form['googlesheets']['issues'] = [
      '#theme' => 'item_list',
      '#list_type' => 'ul',
      '#items' => [],
    ];
    if ($this->configuration['google_api_credential_type'] == 'google_api_service_client') {
      if (!empty($google_api_service_clients[$this->configuration['google_api_service_client']])) {
        /** @var GoogleApiServiceClient $google_api_service_client */
        $google_api_service_client = $google_api_service_clients[$this->configuration['google_api_service_client']];
        $scopes = $google_api_service_client->getScopes();
        $services = $google_api_service_client->getServices();
        $label = $google_api_service_client->label();
      }
    } else {
      if (!empty($google_api_clients[$this->configuration['google_api_client']])) {
        /** @var GoogleApiClient $google_api_client */
        $google_api_client = $google_api_clients[$this->configuration['google_api_client']];
        $scopes = $google_api_client->getScopes();
        $services = $google_api_client->getServices();
        $label = $google_api_client->label();
      }
    }
    // Check if Google Sheets selected as a service.
    if (isset($services) && !in_array('sheets', $services)) {
      $form['googlesheets']['issues']['#items'][] = new TranslatableMarkup('Select %service as a service.', ['%service' => 'Google Sheets API']);
    }
    // Check if Google Sheets selected as a scope.
    if (isset($scopes) && !in_array('https://www.googleapis.com/auth/spreadsheets', $scopes) && !in_array('SPREADSHEETS', $scopes)) {
      $form['googlesheets']['issues']['#items'][] = new TranslatableMarkup('Select %scope as a scope.', ['%scope' => 'https://www.googleapis.com/auth/spreadsheets']);
    }
    // Check if Google API authenticated.
    /** @var GoogleApiClient $google_api_client */
    if (isset($google_api_client) && !$google_api_client->getAuthenticated()) {
      $link = Link::fromTextAndUrl(
        new TranslatableMarkup('here'),
        Url::fromRoute('google_api_client.callback', [], [
          'attributes' => ['target' => '_blank'],
          'query' => ['id' => $google_api_client->id()]
        ])
      );
      $form['googlesheets']['issues']['#items'][] = new TranslatableMarkup('Perform authentication @link', ['@link' => $link->toString()]);
    }

    // Hide the issues list if there are no issues.
    $form['googlesheets']['issues']['#access'] = !empty($form['googlesheets']['issues']['#items']);
    $form['googlesheets']['issues']['#title'] = !empty($form['googlesheets']['issues']['#items'])
      ? new TranslatableMarkup('%label has following issues:', ['%label' => $label])
      : new TranslatableMarkup('%label is correctly configured', ['%label' => $label]);

    // Settings for Multivalued fields.
    $form['multivalue'] = [
      '#type' => 'details',
      '#title' => $this->t('Multivalue fields'),
    ];
    $form['multivalue']['multivalue_limit'] = [
      '#type' => 'number',
      '#title' => $this->t('Number of values'),
      '#description' => $this->t('Limits the number of values sent to Google Sheets for fields with multiple values. For any field, if this limit is greater than the allowed number of values, the allowed number will be used instead.'),
      '#min' => 0,
      '#max' => 1000,
      '#step' => 1,
      '#default_value' => $this->configuration['multivalue_limit'] ?? 0,
    ];
    $form['multivalue']['multivalue_empty'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include empty'),
      '#description' => $this->t('Fields left empty by the users will be included for multivalued fields, limited by %multivalue_limit setting. This setting helps ensure columns always align on future submissions.', ['%multivalue_limit' => $this->t('Number of values')]),
      '#default_value' => $this->configuration['multivalue_empty'] ?? 0,
    ];

    // Submission data.
    $form['submission_data'] = [
      '#type' => 'details',
      '#title' => $this->t('Submission data'),
    ];
    $form['submission_data']['sort_data'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Add submission data columns first'),
      '#description' => $this->t('If checked, webform submission data columns (i.e. submission ID, etc.) will appear in the spreadsheet before the data columns submitted by user.'),
      '#default_value' => $this->configuration['sort_data'] ?? 0,
    ];
    $form['submission_data']['included_data'] = [
      '#type' => 'webform_excluded_columns',
      '#title' => $this->t('Submitted data'),
      '#title_display' => 'invisible',
      '#webform_id' => $this->getWebform()->id(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['included_data'] ?? [],
    ];

    return $this->setSettingsParents($form);
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);

    if (!$form_state->isValueEmpty('spreadsheet_url')) {
      if (preg_match('#https?://docs.google.com/spreadsheets/d/(.+)/edit(\#gid=(\d+))?#', $form_state->getValue('spreadsheet_url'), $matches)) {
        $form_state->setValue('spreadsheet_id', $matches[1]);
        $form_state->setValue('spreadsheet_sheet_id', !empty($matches[3]) ? $matches[3] : 0);
      } else {
        $form_state->setErrorByName('settings][spreadsheet_url', $this->t('Please provide a valid Google Sheet URL.'));
      }
    }

    if ($form_state->getValue('google_api_credential_type') == 'google_api_client' && $form_state->isValueEmpty('google_api_client')) {
      $form_state->setErrorByName('settings][google_api_client', $this->t('Please select Google API OAuth 2.0 Client.'));
    }
    if ($form_state->getValue('google_api_credential_type') == 'google_api_service_client' && $form_state->isValueEmpty('google_api_service_client')) {
      $form_state->setErrorByName('settings][google_api_service_client', $this->t('Please select Google API Service Account.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    // If update operation, do nothing.
    if ($update) {
      return;
    }

    // Prepare request.
    $data = $this->getData($webform_submission);
    $rows = $this->prepareRows($data);
    $request = $this->prepareRequest($this->configuration['spreadsheet_sheet_id'], $rows);

    // Execute request.
    $error = FALSE;
    try {
      $this
        ->makeRequest($this->configuration['spreadsheet_id'], $request)
        ->valid();
    }
    catch (\Exception $e) {
      $error = $e->getMessage();
    }

    // Log detailed message to the 'webform_submission' log.
    $context = [
      'link' => ($webform_submission->id()) ? $webform_submission->toLink($this->t('View'))->toString() : NULL,
      'webform_submission' => $webform_submission,
      'handler_id' => $this->getHandlerId(),
      'operation' => 'google sheets append',
      '%error' => $error,
    ];
    if (!$error) {
      $this->getLogger('webform_submission')->notice("Webform submission had been successfully appended to Google Sheets.", $context);
    }
    else {
      $this->getLogger('webform_submission')->error("%error error occurred while appending submission to Google Sheets.", $context);
    }
  }

  /**
   * Helper method to prepare webform submission fields.
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   A webform submission.
   *
   * @return array
   *   An array with fields.
   */
  private function getData(WebformSubmissionInterface $webform_submission) {
    // Get webform and submission element data.
    $data = $webform_submission->toArray(TRUE);
    $webform = $this->getWebform();

    // Remove unsupported properties from data.
    // These are typically added by other module's like metatag.
    $unsupported_properties = array_combine($this->unsupportedProperties, $this->unsupportedProperties);
    $data = array_diff_key($data, $unsupported_properties);

    // Flatten data and sort data order.
    $element_data = $data['data'];
    unset($data['data']);
    $data = $this->configuration['sort_data']
      ? $data + $element_data
      : $element_data + $data;

    // Included selected submission data.
    $data = array_diff_key($data, $this->configuration['included_data']);

    // Get export configuration options
    $multivalue_limit = $this->configuration['multivalue_limit'];
    $multivalue_empty = $this->configuration['multivalue_empty'];

    // Replace tokens before re-writing composite and multivalued field keys
    $data = $this->replaceTokens($data, $webform_submission);

    // Process values, and flatten all composites and multi-valued fields
    $flatdata = array();
    foreach ($data as $element_key => $value) {
      $element = $webform->getElement($element_key);

      // Webform metadata
      if (!$element) {
        $flatdata[$element_key] = $value;
        continue;
      }

      // Process single-valued fields and continue
      if (!is_array($value)) {
        $flatdata[$element_key] = $this->processValue($element, $value);
        continue;
      }

      // Determine how many multi-values to include by casting #webform_multiple to integer
      $multiple = empty($element['#webform_multiple']) ? 0 : (int)$element['#webform_multiple'];
      // Change unlimited from 1 to 9999 for ease of logic
      $multiple = ($multiple == 1) ? 9999 : $multiple;

      // Single-valued and unlimited custom composites both return true for #webform_multiple
      // Single-valued should return 0 like other element types, but for now identify them with #multiple
      // Only single-valued composites have a #multiple value of false, unlimited have #multiple undefined
      if (($element['#type'] == 'webform_custom_composite') && isset($element['#multiple']) && ($element['#multiple'] == 0)) {
        $multiple = 0;
      }

      // Set $max columns to the lesser of the max allowed values, or the limit configured on the handler
      $max = ($multiple > $multivalue_limit) ? $multivalue_limit : $multiple;

      // Process composites
      if ($element['#webform_composite']) {
        $fields = $element['#webform_composite_elements'];
        // Process a single instance composite
        if ($multiple == 0) {
          foreach ($fields as $field => $composite_element) {
            if (isset($data[$element_key][$field])) {
              $flatdata[$element_key . '_' . $field] = $this->processValue($composite_element, $value[$field]);
            }
            // Custom composites values are nested one level deeper
            elseif (isset($data[$element_key][0][$field])) {
              $flatdata[$element_key . '_' . $field] = $this->processValue($composite_element, $value[0][$field]);
            }
            else {
              $flatdata[$element_key . '_' . $field] = null;
            }
          }
        }
        // Process composites with multiple values
        else {
          // Add user submitted composites up to the multivalue limit set by user.
          $size = (count($value) > $multivalue_limit) ? $multivalue_limit : count($value);
          for ($i = 0; $i < $size; $i++) {
            foreach ($fields as $field => $composite_element) {
              if (isset($data[$element_key][$i][$field])) {
                $flatdata[$element_key . '_' . $field . '_' . $i] = $this->processValue($composite_element, $value[$i][$field]);
              }
              else {
                $flatdata[$element_key . '_' . $field . '_' . $i] = null;
              }
            }
          }
          // Add empty composites up the limit.
          if ($multivalue_empty) {
            for ($size; $size < $max; $size++) {
              foreach ($fields as $field => $composite_element) {
                $flatdata[$element_key . '_' . $field . '_' . $size] = null;
              }
            }
          }
        }
        continue;
      }

      // Process multivalued fields that are not composites
      if ($element['#webform_multiple']) {
        for ($i = 0; $i < $max; $i++) {
          if (isset($data[$element_key][$i])) {
            $flatdata[$element_key . '_' . $i] = $this->processValue($element, $value[$i]);
          }
          elseif ($multivalue_empty) {
            $flatdata[$element_key . '_' . $i] = null;
          }
        }
      }
    }

    // Return values only.
    return array_values($flatdata);
  }

  /**
   * Process a single value
   *
   * @param array $element
   *   An element.
   * @param mixed $value
   *   The element's value.
   *
   * @return mixed
   *   The element's value cast to formats ready for Google Sheets
   */
  protected function processValue(array $element, $value) {
    // Ignore empty and not equal to zero values.
    // @see https://stackoverflow.com/questions/732979/php-whats-an-alternative-to-empty-where-string-0-is-not-treated-as-empty
    if (empty($value) && $value !== 0 && $value !== '0') {
      return $value;
    }

    // Cast markup to string. This only applies to computed Twig values.
    // @see \Drupal\webform\Element\WebformComputedTwig::computeValue
    if ($value instanceof MarkupInterface) {
      return $value = (string) $value;
    }

    $plugin = $this->elementManager->getElementInstance($element);

    // For files send the URI
    if ($plugin instanceof WebformManagedFileBase) {
      /** @var \Drupal\file\FileInterface $file */
      $file = File::load($value);
      if ($file) {
        // Potential TODO: User option to instead return file name with $file->getFilename()
        $value = $file->getFileUri();
      }
    }

    return $value;
  }

  /**
   * Helper method to prepare cell object.
   *
   * @param string $data
   *   A string to insert into a cell.
   *
   * @return CellData
   *   Cells to append in the row.
   */
  private function prepareCell($data) {
    $value = new ExtendedValue();

    switch (gettype($data)) {
      case 'boolean':
        $value->setBoolValue($data);
        break;

      case 'integer':
      case 'double':
        $value->setNumberValue($data);
        break;

      case 'string':
        $value->setStringValue($data);
        break;

      default:
        // Set nothing?
        break;
    }

    $cell = new CellData();
    $cell->setUserEnteredValue($value);

    return $cell;
  }

  /**
   * Helper method to prepare row object.
   *
   * @param array $data
   *   An array with row values.
   *
   * @return RowData
   *   Rows to append in the request.
   */
  private function prepareRows(array $data) {
    $row = new RowData();
    $row->setValues(array_map([$this, 'prepareCell'], array_values($data)));

    return $row;
  }

  /**
   * Helper method to prepare Google API request.
   *
   * @param string $sheet_id
   *   A string with Google Spreadsheet Sheet ID.
   * @param RowData $rows
   *   Rows to append in the request.
   *
   * @return BatchUpdateSpreadsheetRequest
   *   Batch update request for a spreadsheet.
   */
  private function prepareRequest($sheet_id, RowData $rows) {
    $append_request = new AppendCellsRequest();
    $append_request->setSheetId($sheet_id);
    $append_request->setRows($rows);
    $append_request->setFields('userEnteredValue');

    $request = new Request();
    $request->setAppendCells($append_request);

    return new BatchUpdateSpreadsheetRequest(['requests' => [$request]]);
  }

  /**
   * Helper method to perform Google API request.
   *
   * @param string $spreadsheet_id
   *   A string with Google Spreadsheet ID.
   * @param BatchUpdateSpreadsheetRequest $request
   *   Batch update request for a spreadsheet.
   *
   * @return BatchUpdateSpreadsheetResponse
   *   Google API response.
   */
  private function makeRequest($spreadsheet_id, BatchUpdateSpreadsheetRequest $request) {
    // Try to load the client.
    if ($this->configuration['google_api_credential_type'] == 'google_api_service_client') {
      $google_api_service_client = $this->getEntityStorage('google_api_service_client')->load(
        $this->configuration['google_api_service_client'] ?? ''
      );
    } else {
      $google_api_client = $this->getEntityStorage('google_api_client')->load(
        $this->configuration['google_api_client'] ?? 0
      );
    }

    // Get the service based on what type of credentials we are using.
    if (!empty($google_api_service_client)) {
      $this->googleApiServiceClient->setGoogleApiClient($google_api_service_client);
      $service = new Sheets($this->googleApiServiceClient->googleClient);
    } else if (!empty($google_api_client)) {
      $this->googleApiClient->setGoogleApiClient($google_api_client);
      $service = new Sheets($this->googleApiClient->googleClient);
    } else {
      throw new Exception(new TranslatableMarkup("Unable to find Google API Entity with type @type and ID @id", [
        '@type' => $this->configuration['google_api_credential_type'],
        '@id' => $this->configuration[$this->configuration['google_api_credential_type']],
      ]), 1);
    }

    // Perform the request.
    return $service->spreadsheets->batchUpdate($spreadsheet_id, $request);
  }
}
